/******************************************************************************
   FILE:
      skiplist_apply.c

   DESCRIPTION:
      Function to apply a user defined function on skiplist.  This is ac-
      complished by walking one node at a time throught the skiplist on the 
      lowest level.

   RETURN VALUES:
      MY_BOOLEAN signifing whether or not function applied successfully.

   PRE-CONDITIONS:
      Valid skiplist and function.

   POST-CONDITIONS:
      Modified data throught skiplist.

   SIDE EFFECTS:
      None.

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               July 30, 1993
      Undergraduate Research Assistant
      Laboratory for Atmospheric and Space Physics
      Campus Box 590
      Boulder, CO 80309 USA
      pitkow@aries.colorado.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef lint
    static char *RCSid = "$Header: /users/as-csap/pitkow/tar/cs2250/tk3/RCS/skiplist_apply.c,v 1.1 1992/03/02 01:18:12 pitkow Exp $";
#endif

#include "skiplist.h"
    
MY_BOOLEAN skiplist_apply( sk, func )
Skip_List *sk;                           /* header of skiplist     */
MY_BOOLEAN   (*func) ();                    /* user supplied function */
{
    Skip_Node *current;                  /* current pointer        */

    /* handle erronous conditions */
    if ( ! sk || ! func || ! sk->first )
        return ( MY_BOOLEAN ) FALSE;

    /* move through list and apply function; if function fails, abort */
    current = sk->first->forward[0];
    while ( current )
        if ( ( MY_BOOLEAN ) func( current->data )  )
           current = current->forward[0];
        else
           return ( MY_BOOLEAN ) FALSE;
    return ( MY_BOOLEAN ) TRUE;
}
