/******************************************************************************
   FILE:
      skiplist_create.c

   DESCRIPTION:
      Creates header for skiplist and initializes fields to default values.

   RETURN VALUES:
      Valid header, else NULL if bogus parametes passed in or memory failure.

   PRE-CONDITIONS:
      Valid probability and compare function.

   POST-CONDITIONS:
      Header set to default values.

   SIDE EFFECTS:
      Memory used and rand() seed set.

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               July 30, 1993
      Undergraduate Research Assistant
      Laboratory for Atmospheric and Space Physics
      Campus Box 590
      Boulder, CO 80309 USA
      pitkow@aries.colorado.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef lint
    static char *RCSid = "$Header: /users/as-csap/pitkow/tar/cs2250/tk3/RCS/skiplist_create.c,v 1.4 1992/03/03 05:10:58 pitkow Exp $";
#endif

#include "skiplist.h"
    
Skip_List *skiplist_create( cmp, prob )
int   (*cmp) ();                        /* pointer to user defined function */
float prob;                             /* probability used to set level    */
{
    Skip_List *sk;
    int        i;                       /* counter variable                 */

    /* if malloc error then abort and return NULL */
    if (!( sk = ( Skip_List * ) malloc( ( unsigned ) sizeof( Skip_List) ) ) )
    {
        perror( "malloc" );
        (void) fprintf( stderr, "\n ERROR! in skiplist_create \n" );
        return( ( Skip_List * ) NULL );
    }

    /* if cmp points no where (is undefined) then return NULL */
    if ( ! cmp )
    {
        (void) fprintf( stderr, "\n ERROR! compare function not defined \n" );
        return( ( Skip_List * ) NULL );
    }
     
    /* if probability is not with in range of 0..1 then return NULL */
    if ( prob < 0.0 || prob > 1.0 )
    {
        (void) fprintf( stderr, "\n ERROR! probability out of range \n");
        return( ( Skip_List * ) NULL );
    }
        
    /* everythings valid, so define header node's fields to default */
    srand( 1 );
    sk->level      = -1;
    sk->length     = 0;
    sk->num_search = 0;
    sk->prob       = prob;
    sk->compare    = cmp;

    /* if malloc error for "first" field then abort and return NULL */
    if (!( sk->first = ( Skip_Node *) malloc( ( unsigned ) ( sizeof( Skip_Node )
        + ( MAX_LEVEL + 1 ) * (unsigned) sizeof( Skip_Node *) ) ) ) )
    {
        perror( "malloc" );
        (void) fprintf( stderr, "\n ERROR! in skiplist_create \n" );
        return( ( Skip_List * ) NULL );
    }

    for( i = 0 ; i <= MAX_LEVEL; i++ )
        sk->first->forward[ i ] = ( Skip_Node * ) NULL;
    return( sk );
}
