/******************************************************************************
   FILE:
      skiplist_delete.c

   DESCRIPTION:
      Function that deletes a given node in skiplist, if present.
      USER MUST FREE DATA!!!

   RETURN VALUES:
      Pointer to node's data, else NULL if not found in skiplist.

   PRE-CONDITIONS:
      Valid skiplist and node.

   POST-CONDITIONS:
      Shortened list. USER MUST FREE DATA!!!

   SIDE EFFECTS:
      Memory freed.      

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               July 30, 1993
      Undergraduate Research Assistant
      Laboratory for Atmospheric and Space Physics
      Campus Box 590
      Boulder, CO 80309 USA
      pitkow@aries.colorado.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef lint
    static char *RCSid = "$Header: /users/as-csap/pitkow/tar/cs2250/tk3/RCS/skiplist_delete.c,v 1.2 1992/03/03 05:10:58 pitkow Exp $";
#endif

#include "skiplist.h"
    
Datum *skiplist_delete( sk, node )
Skip_List *sk;                        /* header of skiplist                 */
Skip_Node *node;                      /* node to be deleted                 */
{
    int level = 0;                    /* current level of skiplist          */
    int temp_level;                   /* temporary level of skiplist        */
    Datum     *temp_data;             /* temporary data store for retrun    */
    Skip_Node *current;               /* current node                       */
    Skip_Node *trail;                 /* node to be freed                   */

    /* handle erronous conditions */
    if ( ! sk || ! node )
        return ( Datum * ) NULL;
  
    temp_level = sk->level;
    /* if node is in the list, update has the information to relink skiplist */
    if ( trail = skiplist_find( sk, node->data ) )
    {
        /* make sure that 1) current points to a vaild address, an address
           which 2) points to the node to be deleted and 3) that we only
           proceed with this alorithm to the heighest level in the skiplist */ 
        while ( (current = sk->update[ level ]) &&                                               current->forward[ level ] == trail && level <= temp_level )
        {
            current->forward[ level ] = trail->forward[ level ];
            level++;
        }

        temp_data = trail->data;
        free( (char*) trail );

        /* possibly, the height of the skiplist was decreased, so find out */
        while( sk->first->forward[ temp_level ] == ( Skip_Node *) NULL 
             && temp_level >= 0 )
             temp_level--;

        sk->level = temp_level;
        sk->length--;
        return ( Datum * ) temp_data; 
    }
    else
        return ( Datum * ) NULL;
}
