/******************************************************************************
   FILE:
      skiplist_destroy.c

   DESCRIPTION:
      Function that frees skiplist and header. DOES NOT FREE DATA!!! 

   RETURN VALUES:
      Void.

   PRE-CONDITIONS:
      Valid list.

   POST-CONDITIONS:
      Freed memory and first pointer pointing to NULL to avoid accidental
      reaccess of skiplist.      

   SIDE EFFECTS:
      Memory freed.

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               July 30, 1993
      Undergraduate Research Assistant
      Laboratory for Atmospheric and Space Physics
      Campus Box 590
      Boulder, CO 80309 USA
      pitkow@aries.colorado.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef lint
    static char *RCSid = "$Header: /users/as-csap/pitkow/tar/cs2250/tk3/RCS/skiplist_destroy.c,v 1.1 1992/03/02 01:18:12 pitkow Exp $";
#endif

#include "skiplist.h"
    
void skiplist_destroy (sk)
Skip_List *sk;                           /* header of skiplist */
{
    Skip_Node *current;                  /* current pointer    */
    Skip_Node *trail;                    /* previous pointer   */

    /* process only if there is a skiplist present */
    if ( sk )
    {   
        /* move through list and free memory, than free header */
        current = sk->first;
        while ( current )
        {
            trail = current->forward[0];
            free( (char*) current );
            current = trail;
        }

        /* prevent accidental reaccess of skiplist */
        sk->first = ( Skip_Node * ) NULL;
        free( (char*) sk );
    }
}
