/******************************************************************************
   FILE:
      skiplist_find.c

   DESCRIPTION:
      Function that searches the skiplist using the user defined compare
      function for the particular data.  The search starts at the heighest 
      level of the list and proceeds to drop down to the next level if the
      target data is greater than the next node of that level being compared.
      The header's update vector keeps track of where this drop occurs.  This
      information becomes usefull in skiplist_delete and skiplist_insert,
      since previous node pointer redirection is necessary.

   RETURN VALUES:
      Address of node if found, else NULL.

   PRE-CONDITIONS:
      Valid list and target data.

   POST-CONDITIONS:
      None.

   SIDE EFFECTS:
      Increments num_searched in header node.

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               July 30, 1993
      Undergraduate Research Assistant
      Laboratory for Atmospheric and Space Physics
      Campus Box 590
      Boulder, CO 80309 USA
      pitkow@aries.colorado.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef lint
    static char *RCSid = "$Header: /users/as-csap/pitkow/tar/cs2250/tk3/RCS/skiplist_find.c,v 1.2 1992/03/02 23:38:33 pitkow Exp $";
#endif

#include "skiplist.h"
    
Skip_Node *skiplist_find( sk, target )
Skip_List *sk;                        /* header of skiplist                 */
Datum     *target;                    /* pointer to data being searched for */
{
    int        level;                 /* current level being processed      */
    int        temp = -1;             /* result of compare function         */
    Skip_Node *current;               /* current pointer                    */
    Skip_Node *trail;                 /* previous pointer                   */

    /* handle erronous conditions */
    if ( ! sk || ! target )
    {
        (void) fprintf( stderr, "\n ERROR! occurred in skiplist_find \n" );
        return ( Skip_Node * ) NULL;
    }

    level   = sk->level;
    trail   = sk->first;
    /* start at the heightest level in the skiplist and check node's data
       against target data; if greater than target, drop down one level and
       set the update vector to the address of the previous node; process until
       the lowest level of the skiplist */
    while ( level >= 0 )
    {
        current = trail->forward[ level ];
        if ( current )
        {
            temp = sk->compare( current->data, target );
            sk->num_search++;
            while ( current && temp < 0 )
            { 
                trail   = current;
                current = trail->forward[ level ];
                if ( current )
                {
                    temp = sk->compare( current->data, target );
                    sk->num_search++;
                }
            }
        }
        sk->update[ level ] = trail;
        level--;
    }
    
    /* if target was found return address of that node, else return NULL */
    if ( temp == 0 )
        return ( Skip_Node * ) current;
    else
        return ( Skip_Node * ) NULL;
}
