/******************************************************************************
   FILE:
      skiplist_insert.c

   DESCRIPTION:
      Function that inserts, if not already existant, inorder, data into list.
      Based upon "prob" in the header, the newly inserted node's height will
      be determined.  At most, the inserted node's height will be one greater
      than the heightest node already in the list.

   RETURN VALUES:
      Address of inserted node, else NULL if errors or duplication.

   PRE-CONDITIONS:
      Valid skiplist and data.

   POST-CONDITIONS:
      Lengthened list which possibly was heightened.

   SIDE EFFECTS:
      Memory used for new node.

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               July 30, 1993
      Undergraduate Research Assistant
      Laboratory for Atmospheric and Space Physics
      Campus Box 590
      Boulder, CO 80309 USA
      pitkow@aries.colorado.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef lint
    static char *RCSid = "$Header: /users/as-csap/pitkow/tar/cs2250/tk3/RCS/skiplist_insert.c,v 1.4 1992/03/03 05:10:58 pitkow Exp $";
#endif

#include "skiplist.h"
    
Skip_Node *skiplist_insert( sk, data )
Skip_List *sk;                          /* header of skiplist             */
Datum     *data;                        /* pointer to user defined data   */
{
    int        level;                   /* keeps track of current level   */
    Skip_Node *new_node;                /* new node to be inserted        */
    Skip_Node *current;                 /* current pointer                */

    level = sk->num_search;

    /* utilize find function; if already in list, return address */
     current = skiplist_find( sk, data );  
     if ( current != NULL ) {
        return ( Skip_Node * ) current;
     }

     sk->num_search = level;

    /* if node was not found, the update vector, via the find function, will 
       contain the location to insert the new node; so generate a random level,
       update the header's and update's level if necessary */
    level = skiplist_genlevel( sk->prob );
    if ( ( level > sk->level ) && ( sk->level <= MAX_LEVEL ) )
    {
         sk->level++;
         level = sk->level;
         /* point new level in update vector to first node in list */
         sk->update[ level ] = sk->first;
    }

    /* handle possible malloc error */
    if ( !( new_node = ( Skip_Node *) malloc( ( unsigned ) (sizeof( Skip_Node )
         + ( level ) * ( unsigned ) sizeof( Skip_Node *) ) ) ) )
    {
        perror( "malloc" );
        (void) fprintf( stderr, "\n ERROR! in skiplist_insert \n" );
        return ( Skip_Node * ) NULL ;
    }

    sk->length++;
    sk->first->height = level;
    new_node->height = level;
    new_node->data   = data;
    /* link new_node to forward nodes and previous nodes to new_node */
    do
    {
        current = sk->update[ level ];
        new_node->forward[ level ] = current->forward[ level ];
        current->forward[ level ]  = new_node;
    }
    while ( --level >= 0 );

    return ( Skip_Node * ) new_node;
}
