/*		Mosaic HTML text object
**		=================
**
**	This version of the stream object just writes to a socket.
**	The socket is assumed open and left open.
**
**	Bugs:
**		strings written must be less than buffer size.
*/
#include "HTMosaicHTML.h"

#define BUFFER_SIZE 4096;	/* Tradeoff */

#include "HTUtils.h"
#include "HText.h"



/*		HTML Object
**		-----------
*/

struct _HTStream {
	CONST HTStreamClass *	isa;

	HText * 		text;
};

/*	Write the buffer out to the socket
**	----------------------------------
*/


/*_________________________________________________________________________
**
**			A C T I O N 	R O U T I N E S
*/

/*	Character handling
**	------------------
*/

PRIVATE void HTMosaicHTML_put_character ARGS2(HTStream *, me, char, c)
{
    HText_appendCharacter(me->text, c);
}



/*	String handling
**	---------------
**
*/
PRIVATE void HTMosaicHTML_put_string ARGS2(HTStream *, me, CONST char*, s)
{
    HText_appendText(me->text, s);
}


PRIVATE void HTMosaicHTML_write ARGS3(HTStream *, me, CONST char*, s, int, l)
{
    CONST char* p;
    CONST char* e = s+l;
    for (p=s; p<e; p++) HText_appendCharacter(me->text, *p);
}



/*	Free an HTML object
**	-------------------
**
**	Note that the SGML parsing context is freed, but the created object is not,
**	as it takes on an existence of its own unless explicitly freed.
*/
PRIVATE void HTMosaicHTML_free ARGS1(HTStream *, me)
{
    free(me);
}

/*	End writing
*/

PRIVATE void HTMosaicHTML_end_document ARGS1(HTStream *, me)
{
    HText_endAppend(me->text);
}



/*		Structured Object Class
**		-----------------------
*/
PUBLIC CONST HTStreamClass HTMosaicHTML =
{		
	"SocketWriter",
	HTMosaicHTML_free,
	HTMosaicHTML_end_document,
	HTMosaicHTML_put_character, 	HTMosaicHTML_put_string, HTMosaicHTML_write,
}; 


/*		New object
**		----------
*/
PUBLIC HTStream* HTMosaicHTMLPresent ARGS3(
	HTPresentation *,	pres,
	HTParentAnchor *,	anchor,	
	HTStream *,		sink)
{

    HTStream* me = (HTStream*)malloc(sizeof(*me));
    if (me == NULL) outofmem(__FILE__, "HTMosaicHTMLPresent");
    me->isa = &HTMosaicHTML;       

    me->text = HText_new(anchor);
    HText_beginAppend(me->text);

    return (HTStream*) me;
}
