/*	HyperText Tranfer Protocol	- Client implementation		HTTP.c
**	==========================
**
** Bugs:
**	Not implemented:
**		Forward
**		Redirection
**		Error handling
*/

#include "HTTP.h"

#define HTTP_VERSION	"HTTP/1.0"

#define CR   FROMASCII('\015')	/* Must be converted to ^M for transmission */
#define LF   FROMASCII('\012')	/* Must be converted to ^J for transmission */

#define INIT_LINE_SIZE		1024	/* Start with line buffer this big */
#define LINE_EXTEND_THRESH	256	/* Minimum read size */
#define VERSION_LENGTH 		20	/* for returned protocol version */

#include "HTParse.h"
#include "HTUtils.h"
#include "tcp.h"
#include "HTTCP.h"
#include "HTFormat.h"
#include "HTFile.h"
#include <ctype.h>
#include "HTAlert.h"
#include "HTMIME.h"
#include "HTML.h"		/* SCW */
#include "HTInit.h"		/* SCW */

/* #define TRACE 1 */

struct _HTStream {
	HTStreamClass * isa;		/* all we need to know */
};


extern char * HTAppName;	/* Application name: please supply */
extern char * HTAppVersion;	/* Application version: please supply */

/*		Load Document from HTTP Server			HTLoadHTTP()
**		==============================
**
**	Given a hypertext address, this routine loads a document.
**
**
** On entry,
**	arg	is the hypertext reference of the article to be loaded.
**
** On exit,
**	returns	>=0	If no error, a good socket number
**		<0	Error.
**
**	The socket must be closed by the caller after the document has been
**	read.
**
*/
PUBLIC int HTLoadHTTP ARGS4 (
	CONST char *, 		arg,
	HTParentAnchor *,	anAnchor,
	HTFormat,		format_out,
	HTStream*,		sink)
{
  int s;				/* Socket number for returned data */
  char *command;			/* The whole command */
  char * eol = 0;			/* End of line if found */
  char * start_of_data;		/* Start of body of reply */
  int status;				/* tcp return */
  int bytes_already_read = 0;
  char crlf[3];			/* A CR LF equivalent string */
  HTStream *	target = NULL;		/* Unconverted data */
  HTFormat format_in;			/* Format arriving in the message */
  
  BOOL had_header = NO;		/* Have we had at least one header? */
  char * line_buffer = NULL;
  char * line_kept_clean = NULL;
  BOOL extensions = YES;		/* Assume good HTTP server */
  
  int length = 0;
  int recycle_first_chunk = 0;
  
  if (!arg) return -3;		/* Bad if no name sepcified	*/
  if (!*arg) return -2;		/* Bad if name had zero length	*/
  
  sprintf(crlf, "%c%c", CR, LF);	/* To be corect on Mac, VM, etc */

	/* added (char *) dave brown;  brownd@csc.ac.liv.uk  */ 
  status = HTDoConnect ((char *) arg, "HTTP", TCP_PORT, &s);
  if (status < 0) 
    {
      if (TRACE) 
        fprintf(stderr, 
                "HTTP: Unable to connect to remote host for `%s' (errno = %d).\n", arg, errno);
      HTProgress ("Unable to connect to remote host.");
      return HT_NO_DATA;
    }
  
  /*	Ask that node for the document,
   **	omitting the host name & anchor
   */        
  {
    char * p1 = HTParse(arg, "", PARSE_PATH|PARSE_PUNCTUATION);
    command = malloc(4 + strlen(p1)+ 2 + 31);
    if (command == NULL) outofmem(__FILE__, "HTLoadHTTP");
    strcpy(command, "GET ");
    strcat(command, p1);
    free(p1);
  }
  if (extensions) 
    {
      strcat(command, " ");
      strcat(command, HTTP_VERSION);
    }
  
  strcat(command, crlf);	/* CR LF, as in rfc 977 */
  
  if (extensions) 
    {
      int n;
      int i;
      HTAtom * present = WWW_PRESENT;
      char line[256];    /*@@@@ */
      
      if (!HTPresentations) HTFormatInit();
      n = HTList_count(HTPresentations);
      
      for(i=0; i<n; i++) 
        {
          HTPresentation * pres = HTList_objectAt(HTPresentations, i);
          if (pres->rep_out == present) 
            {
              if (pres->quality != 1.0) 
                {
                  sprintf(line, "Accept: %s q=%.3f%c%c",
                          HTAtom_name(pres->rep), pres->quality, CR, LF);
                } 
              else 
                {
                  sprintf(line, "Accept: %s%c%c",
                          HTAtom_name(pres->rep), CR, LF);
                }
              StrAllocCat(command, line);
            }
        }
      
      sprintf(line, "User-Agent:  %s/%s  libwww/%s%c%c",
              HTAppName ? HTAppName : "unknown",
              HTAppVersion ? HTAppVersion : "0.0",
              HTLibraryVersion, CR, LF);
      StrAllocCat(command, line);
    }
  
  StrAllocCat(command, crlf);	/* Blank line means "end" */
  
#if 0
  if (TRACE) 
    fprintf(stderr, "HTTP Tx: %s\n", command);
#endif  

  status = NETWRITE(s, command, (int)strlen(command));
  free(command);
  if (status<0) 
    {
      if (TRACE) fprintf(stderr, "HTTPAccess: Unable to send command.\n");
      HTProgress ("Unable to send command to remote host.");
      return HT_NO_DATA;
    }
  
  /*	Read the first line of the response
   **	-----------------------------------
   */
  
  {
    /* Get numeric status etc */
    
    BOOL end_of_file = NO;
    HTAtom * encoding = HTAtom_for("8bit");
    int buffer_length = INIT_LINE_SIZE;	/* Why not? */
    
    line_buffer = (char *) malloc(buffer_length * sizeof(char));
    if (!line_buffer) outofmem(__FILE__, "HTLoadHTTP");
    
    do 
      {	/* Loop to read in the first line */
        /* Extend line buffer if necessary for those crazy WAIS URLs ;-) */
        
        if (buffer_length - length < LINE_EXTEND_THRESH) 
          {
            buffer_length = buffer_length + buffer_length;
            line_buffer = 
              (char *) realloc(line_buffer, buffer_length * sizeof(char));
            if (!line_buffer) outofmem(__FILE__, "HTLoadHTTP");
          }
        status = NETREAD(s, line_buffer + length,
                         buffer_length - length -1);
        if (status < 0) 
          {
            HTAlert("Unexpected network read error; connection aborted.");
            NETCLOSE(s);
            return status;
          }
        
#if 0
        if (TRACE) fprintf(stderr, "HTTP: read returned %d bytes.\n",
                           status);
#endif
        bytes_already_read += status;
        {
          char line[256];
          sprintf (line, "Read %d bytes of data.", status);
          HTProgress (line);
        }
        
        if (status == 0) 
          {
            end_of_file = YES;
            break;
          }
        line_buffer[length+status] = 0;
        
        if (line_buffer)
          {
            if (line_kept_clean)
              free (line_kept_clean);
            line_kept_clean = (char *)malloc (buffer_length * sizeof (char));
            bcopy (line_buffer, line_kept_clean, buffer_length);
          }

        eol = strchr(line_buffer + length, LF);
        if (eol && *(eol-1) == CR) 
          *(eol-1) = ' '; 
        
        length = length + status;
        
        if (eol) 
          *eol = 0;		/* Terminate the line */
      } 
    while (!eol && !end_of_file);		/* No LF */	    
  } /* Scope of loop variables */
    
    
  /*	We now have a terminated unfolded line. Parse it.
   **	-------------------------------------------------
   */
  if (TRACE)
    fprintf(stderr, "HTTP: Rx: %s\n", line_buffer);
  
  {
    int fields;
    char server_version[VERSION_LENGTH+1];
    int server_status;

    server_version[0] = 0;
    
    fields = sscanf(line_buffer, "%20s %d",
                    server_version,
                    &server_status);
    
    if (TRACE)
      fprintf (stderr, "HTTP: Scanned %d fields from line_buffer\n", fields);
    
    /* Rule out HTTP/1.0 reply as best we can. */
    if (fields < 2 || !server_version[0] || server_version[0] != 'H' ||
        server_version[1] != 'T' || server_version[2] != 'T' ||
        server_version[3] != 'P' || server_version[4] != '/' ||
        server_version[6] != '.') 
      {			/* HTTP0 reply */
        HTAtom * encoding;

        if (TRACE)
          fprintf (stderr, "--- Talking HTTP0.\n");
        	/* added (char *) dave brown;  brownd@csc.ac.liv.uk  */ 
        format_in = HTFileFormat((char *) arg, &encoding, WWW_HTML);
        start_of_data = line_kept_clean;
        recycle_first_chunk = 1;
      } 
    else 
      {
        /*	Decode full HTTP response */
        format_in = HTAtom_for("www/mime");
        start_of_data = eol ? eol + 1 : "";

        if (TRACE)
          fprintf (stderr, "--- Talking HTTP1.\n");
        
        switch (server_status / 100) 
          {
          default:		/* bad number */
            HTAlert("Unknown status reply from server!");
            break;
            
          case 3:		/* Various forms of redirection */
            HTAlert("Redirection response from server is not handled by this client");
            break;
            
          case 4:		/* "I think I goofed" */
          case 5:		/* I think you goofed */
            {
              char *p1 = HTParse(arg, "", PARSE_HOST);
              char * message = (char*)malloc(strlen(line_buffer)+strlen(p1) + 100);
              if (!message) outofmem(__FILE__, "HTTP 5xx status");
              sprintf(message,
                      "HTTP server at %s replies:\n%s", p1, line_buffer);
              status = HTLoadError(sink, server_status, message);
              free(message);
              free(p1);
              goto clean_up;
            }
            break;
            
          case 2:		/* Good: Got MIME object */
            break;
            
          } /* switch on response code */
      }	/* Full HTTP reply */
  } /* scope of fields */
  
  /* Set up the stream stack to handle the body of the message */
  target = HTStreamStack(format_in,
                         format_out,
                         sink, anAnchor);
  
  if (!target) 
    {
      char buffer[1024];	/* @@@@@@@@ */
      if (line_buffer) free(line_buffer);
      if (line_kept_clean) free(line_kept_clean);
      sprintf(buffer, "Sorry, no known way of converting %s to %s.",
              HTAtom_name(format_in), HTAtom_name(format_out));
      fprintf(stderr, "HTTP: %s", buffer);
      status = HTLoadError(sink, 501, buffer);
      goto clean_up;
    }

  if (recycle_first_chunk)
    {
      if (TRACE)
        fprintf (stderr, "[HTTP] Recycling first chunk of data, length %d\n", length);
      (*target->isa->put_block)(target, start_of_data, length);
    }
  else
    {
      (*target->isa->put_string)(target, start_of_data);
    }

  /* Go pull the bulk of the data down. */
  HTCopy(s, target, bytes_already_read);
  
  (*target->isa->end_document)(target);
  (*target->isa->free)(target);
  status = HT_LOADED;
  
  /*	Clean up
   */
  
 clean_up: 
  if (line_buffer) free(line_buffer);
  if (line_kept_clean) free(line_kept_clean);
  
  (void) NETCLOSE(s);
  
  return status;			/* Good return */
}


/*	Protocol descriptor
*/

PUBLIC HTProtocol HTTP = { "http", HTLoadHTTP, 0 };
