/****************************************************************************** 
   FILE:
      html_analyzer.c

   DESCRIPTION:
      Reads Hypertext Markup Language (HTML) documents, extracts 
      information to check availability of documents pointed to
      by HTML anchor entity (except telent, rlogin), writes non-html 
      version of file to temporary file for future processing, performs
      the desired tests ( verify, completeness, consistent_content,
      consistent_link).

      Global varaibles used:
        extern char 	*respostiory;

      Internal variables used:
        char     	non_html_file[BUFSIZ];
        FILE    	*non_html_file_out;
	Skip_List    	*skiplist1;             
	Skip_List    	*skiplist2;             

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               October 7, 1993
      Graphics, Visualization and Usability Laboratory
      Georgia Institute of Technology
      pitkow@cc.gatech.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/
#include "html_analyzer.h"

/*
 * global declarations
 */
extern char *repository;   	/* where to put link info &  non_html files */


/*
 * internal variable declarations
 */
char 	 non_html_file[BUFSIZ];	/* the name of the file being processed */
FILE	*non_html_file_out;	/* file pointer to non_html file        */
Skip_List    *skiplist1;        /* skiplist ordered by links */
Skip_List    *skiplist2;        /* skiplist ordered by content  */

char *get_href ( orig_anchor)
char *orig_anchor;
{
	/*  added "=NULL"  1/3/95 ; dave brown (brownd@csc.ac.liv.uk */
   char *href=NULL;
   char *anchor=NULL;
   int   quoted = 0;

   /*
    * double check for non null anchor
    */
   if ( !strlen( orig_anchor) || !orig_anchor)
      return (char *)NULL;

   if ( !( href = ( char *) malloc ( sizeof( char) * (strlen( anchor) + 1)))) {
      fprintf( stderr, "\nget_href: malloc error");
      safe_bail_out( -1);
   }
   bzero( href, (strlen( anchor) + 1));

   /*
    * make copy of original anchor
    */
   anchor = strdup( orig_anchor);


	/* patch-1 added */
     /* make copy of original anchor
      */
     anchor = strdup( orig_anchor);

 /* Here's the  patched bit */ 
 
   if ( !anchor) {
       fprintf( stderr, "\nget_href: malloc error");
       safe_bail_out( -1);
    }
  /* End patch */

  
   /*
    * handle non-hyperlink anchors , e.g. <A NAME=1</A>
    */ 
   if ( ! ( anchor = ( char *)strstr( anchor, "HREF")) && 
        ! ( anchor = ( char *)strstr( anchor, "href")) )
      return " \0";
 
   /*
    * attempt to get the value of the href
    */
   while ( anchor) {
      if ( *++anchor == '=') {

         /*
          * treat quoted values differently
          */
         if ( *++anchor == '"') {
            quoted = 1;

            /*
             * return if HREF="" or HREF=" " 
             */
            if ( *++anchor == '"' ||
                 *( anchor + 1) == '"') 
               return (char *)NULL;
         }

         /* 
          * return if HREF="#foo" or HREF=#foo or if the access
          * method is either telnet or rlogin, since these require
          * an interactive user and do not return documents
          */
         if ( *anchor == '#' || ( strncasecmp( anchor, "telnet", 6) == 0)
                             || ( strncasecmp( anchor, "rlogin", 6) == 0))
            return (char *)NULL;

         /* 
          * copy the value of the href, and return 
          */
         href = anchor;
         if ( quoted) { 
            int	index = strlen( href);

            /*
             *  parse back until trailing quote is found
             */
            while ( href)  
               if( href[ index] == '"') {
                  href[ index] = '\0';
                  break;
               }
               else
                  href[ index--] = '\0';
         }
         return( href);
      }
   }
   return (char *)NULL;
}

extract_links( file, size)
char	*file;
int	 size;
{
   struct mark_up *tag_list;
   FILE		*in_file;
   Skip_Data	*data1;
   Skip_Data	*data2;
   Skip_Node    *node;                  
   char			*buffer, *href1, *prog = "extract-links";


   /*
    * open the input file
    */
   if ( !( in_file = fopen( file, "r"))) {
      fprintf( stderr, "\n%s: can not open <%s>\n", prog, file);
      safe_bail_out( -2);
   }

	/* inserted "non_html_file" in output string; D.M.Brown@csc.liv.ac.uk */
   if ( !( non_html_file_out = fopen( non_html_file, "w"))) {
      fprintf( stderr, "\n%s: can not open non_html_file%s", prog, non_html_file);
      safe_bail_out( -2);
   }

   /*
    * malloc enought space for the contents fo the file and read in
    */
   if ( !( buffer = malloc ( sizeof ( char) * size))) {
      fprintf( stderr, "\n%s: , malloc error", prog);
      safe_bail_out( -3);
   }

   if ( !fread( buffer, sizeof( char), size, in_file)) {
            fprintf( stderr, "\n%s: can not read %s", prog, file);
      safe_bail_out( -2);
   }

   /*
    * use HTMLParse to build a list of all tags from the file
    */
   tag_list = (struct mark_up*) HTMLParse((struct mark_up*)NULL, (char*) buffer); 

   /*
    * loop through the list adding anchors to the skiplist and creating
    * a non-html version of the file
    */
   while ( tag_list) {

      /*
       * do not add the contents of the title to the non-html file
       */ 
      if ( tag_list->type == M_TITLE) 
         for ( tag_list = tag_list->next; 
               tag_list && !tag_list->is_end && tag_list->type != M_TITLE; 
               tag_list = tag_list->next);
      if ( !tag_list) 
         break;

      /*
       * allow anchors to have embedded tags, but look for images  
       */
      if ( tag_list->type == M_ANCHOR && !tag_list->is_end) {
          
         href1 = get_href( tag_list->start);

         /*
          * some anchors are null, others point to locations within
          * the same file; either way, move on to next tag
          */
         if ( href1 == NULL) {
            for ( tag_list = tag_list->next;
               tag_list && !tag_list->is_end && tag_list->type != M_ANCHOR;
               tag_list = tag_list->next);
            tag_list = tag_list->next;
            continue;
         }

         /*
          * get a new data pointers 
          */
         data1 = init_data();
         data2 = init_data();
         strcpy( data1->file_name, file);

         /*
          * copy href into data struct and move to inner most tag
          */
         strcpy( data1->href, href1);
         free( href1);
         while ( tag_list && tag_list->type != M_NONE && tag_list->type != M_IMAGE)
            tag_list = tag_list->next;

         /* 
          * if the tag is an image use the start field, else use the 
          * text field as the anchor's content
          */ 
         if ( !tag_list)
            break;
         else if ( tag_list->type == M_NONE) 
            strcpy( data1->content, tag_list->text);
         else if ( tag_list->type == M_IMAGE)
            strcpy( data1->content, tag_list->start);
         else {
            fprintf( stderr, "%s: unable to find content of anchor\n", prog);
            break;
         }
         
         /*
          * move through list until the end of the anchor
          */
         while ( tag_list && tag_list->type != M_ANCHOR)
            tag_list = tag_list->next;


         /* 
          * make a copy of the data for the skiplist ordered by anchor
          * address and the other by anchor contents
          */
         copy_data( data1, data2);

         /* 
          * insert into skiplists, using differnt compare functions and
          * incrementing the appropriate occurrence field
          */
         if ( !( node = skiplist_insert( skiplist1, (Datum *) data1))) {
            fprintf( stderr, "\n%s: Skiplist_insert failed", prog);
            safe_bail_out( -3);
         }
         else
            ( ( Skip_Data *)node->data)->occurences++;

         if ( !( node = skiplist_insert( skiplist2, (Datum *) data2))) {
            fprintf( stderr, "\n%s: Skiplist_insert failed", prog);
            safe_bail_out( -3);
         }
         else
            ( ( Skip_Data *)node->data)->occurences++;

         /*
          * get a new data pointers since the last are now in the skiplists
          */
         data1 = init_data();
         data2 = init_data();
         strcpy( data1->file_name, file); 
      }
 
      /*
       * if is possible that the above special cases left the pointer
       * pointing to non html text, so output text to file
       */
      if ( tag_list->type == M_NONE ) 
         fprintf( non_html_file_out, "%s", tag_list->text);

      tag_list = tag_list->next;
   }

   fclose( in_file);
   fclose( non_html_file_out);
}

void html_analyzer( directory, supress)
char	*directory;
short    supress;
{
   char         *prog = "html_analyzer";

   /* 
    * set up skiplists
    */
   if ( !( skiplist1 = ( Skip_List *) skiplist_create( cmp_data_link, 0.5))) {
      fprintf( stderr, "\n%s: Skiplist_Create failed", prog);
      safe_bail_out( -3);
   }

   if ( !( skiplist2 = ( Skip_List *) skiplist_create( cmp_data_content, 0.5))) {
      fprintf( stderr, "\n%s: Skiplist_Create failed", prog);
      safe_bail_out( -3);
   }

   /*
    * call resurce function to load skiplist with hrefs and create
    * non-html files
    */
    recurse_fcn( directory, "html");

   /*
    * only perform the specified tests
    */
   if ( !( supress & VALIDATE)) {
      printf( "\nVERIFYING LINKS...\n\n");
      skiplist_apply( skiplist1, validate);
      printf( "Done Verifying links.\n");
   }

   if ( !(supress & COMPLETE)) {
      printf( "\nVERIFYING COMPLETENESS...\n\n");
      skiplist_apply( skiplist1, completeness);
      printf( "Done Verifying completeness.\n");
   }

   if ( !(supress & CONSISTENT)) {
      printf( "\nVERIFYING CONSISTENCY OF LINKS...\n\n");
      consistent_link( skiplist1);
      printf( "Done Verifying consistency of links.\n");
      printf( "\nVERIFYING CONSISTENCY OF CONTENTS...\n\n");
      consistent_content( skiplist2);
      printf( "Done Verifying consistency of contents.\n");
   }

   skiplist_apply( skiplist1, free);
   skiplist_apply( skiplist2, free);
   skiplist_destroy( skiplist1);
   skiplist_destroy( skiplist2);
}

recurse_fcn ( name, extension)
char    *name;
char    *extension;
{
   struct stat   statbuffer;
   FILE         *fileptr;
   REG int	 index	= 0,
		 name_index	= strlen( name),
		 extn_index	= strlen( extension);

   if ( stat( name, &statbuffer) == -1) {
      fprintf( stderr, "\nrecurse_fcn: can not access %s", name);
      safe_bail_out( -2);
   }

   if ( ( statbuffer.st_mode & S_IFMT) == S_IFDIR) {
      printf( "\nProcessing files in directory: %s...", name);
      dir_recurse( name, extension);
      return;
   }
   
   /*
    * only process filenames that end with the contents of extension 
    */
   while( index <= name_index && index <= extn_index ) 
      if ( extension[ extn_index - index] != name[ name_index - index])
         return;
      else
         index++;

   /* 
    * process file
    */
   extract_links( name, statbuffer.st_size); 
}

dir_recurse ( dir, extension)
char    *dir;
char    *extension;
{
   char          *name;
   DIR           *dir_file_desc;
   struct dirent *dir_pointer;

   if ( ( dir_file_desc = opendir( dir)) == NULL) {
      fprintf( stderr, "\ndir_recurse: can not open %s", dir);
      safe_bail_out (-2);
   }

   while ( dir_pointer = readdir( dir_file_desc)) {
      int length = strlen( dir) + strlen( dir_pointer->d_name) + 2;

      if ( !strcmp( dir_pointer->d_name, ".") ||
           !strcmp( dir_pointer->d_name, ".."))
         continue;

      /*
       * make sure the full path name will fit into name 
       */
      if ( !( name = malloc ( sizeof ( char) * length))) {
         fprintf( stderr, "\ndir_recurse: , malloc error");
         safe_bail_out( -3);
      }

      /* 
       * set the name of the non_html file while the info is handy
       */
      bzero( name, length);
      non_html_file[0] = '\0';
      sprintf( name, "%s/%s", dir, dir_pointer->d_name);
      sprintf( non_html_file, "%s/%s", repository, dir_pointer->d_name);
      recurse_fcn( name, extension);
      free( name);
   }
   closedir( dir_file_desc);
}
