/******************************************************************************
   FILE:
      html_analyzer.h

   DESCRIPTION:
      Contains declarations for the html_analyzer. 

   DEVELOPED BY:                                 DATE:
      James E. Pitkow                               October 7, 1993
      Graphics, Visualization and Usability Laboratory
      Georgia Institute of Technology
      pitkow@cc.gatech.edu

   COPYRIGHT:
      The Regents of the University of Colorado, 1993

******************************************************************************/

#ifndef HTML_ANALYSER_H
#define HTML_ANALYSER_H

/*
 * both necessary to avoid compiliation errors with
 * other libraries
 */
#define USE_DIRENT
#define _NO_PROTO

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h> 
#include <malloc.h>
#include <fcntl.h>
#include "../libhtmlw/HTMLP.h"
#include "../libskiplist/skiplist.h"


#define VALIDATE        01	/* flag to supress validate option     */
#define COMPLETE        02	/* flag to supress completeness option */
#define CONSISTENT      04	/* flag to supress consistency option  */

#define REG	register

typedef struct skip_data {     
    int         occurences;       /* number of times the link occurs w/a content */
    char        href[BUFSIZ];     /* the html HREF link address              */
    char        content[BUFSIZ];  /* the href identifier                     */
    char        file_name[BUFSIZ];/* the file name where href extracted from */
} Skip_Data;


void    	html_analyzer();  /* extracts links, and performs tests  */
int  		HTLoadHTTP();     /* verifies existence of href          */ 
int             cmp_data_link();  /* compare based on href value         */
int             cmp_data_content(); /* compare based on content of link */
Skip_Data       *init_data();	  /* initialize data struct for sklist   */
MY_BOOLEAN      validate();       /* verify availability of documents    */
MY_BOOLEAN      completeness();   /* find incomplete use of link content */
void            consistent_link();/* find links with many link contents  */
void            consistent_content(); /* find contents with many links   */


#endif
